@@ -24,6 +24,7 @@ class GPServer(object):
         self.url = url
         self.username = username
         self.password = password
+        self.token = None
         self.last_job = None
 
     def __str__(self):
@@ -53,7 +54,8 @@ class GPServer(object):
         response = urllib.request.urlopen(request, b'')
         if response.getcode() != 200:
             raise urllib.error.HTTPError(url, response.getcode(), 'Invalid username or password', response.getheaders(), None)
-        return json.loads(response.read())['access_token']
+        self.token = json.loads(response.read())['access_token']
+        return self.token
 
     def upload_file(self, file_name, file_path):
         """
@@ -130,6 +132,11 @@ class GPServer(object):
             job.wait_until_done()
         return job
 
+    def get_token(self):
+        """Return the authentication token, logging in to obtain it if necessary"""
+        if self.token: return self.token
+        else: return self.login()
+
     def get_job(self, job_number):
         job = GPJob(self, job_number)
         return job
@@ -251,7 +258,13 @@ class GPFile(GPResource):
         if self.server_data.authorization_header() is not None:
             request.add_header('Authorization', self.server_data.authorization_header())
         request.add_header('User-Agent', 'GenePatternRest')
-        return urllib.request.urlopen(request)
+        try:
+            return urllib.request.urlopen(request)
+        except urllib.error.HTTPError as e:
+            if e.geturl():  # Handle S3 redirects if one is encountered
+                return urllib.request.urlopen(urllib.request.Request(e.geturl()))
+            else:
+                raise e
 
     def read(self):
         """
@@ -297,6 +310,7 @@ class GPJob(GPResource):
     output_files = None
     num_output_files = None
     children = None
+    input_params = None
 
     def __init__(self, server_data, uri):
         super(GPJob, self).__init__(str(uri))
@@ -318,7 +332,7 @@ class GPJob(GPResource):
             * URL of Output Files
             * Number of Output Files
         """
-        request = urllib.request.Request(self.server_data.url + "/rest/v1/jobs/" + self.uri)
+        request = urllib.request.Request(self.server_data.url + "/rest/v1/jobs/" + self.uri + "?includeInputParams=true")
         if self.server_data.authorization_header() is not None:
             request.add_header('Authorization', self.server_data.authorization_header())
         request.add_header('User-Agent', 'GenePatternRest')
@@ -344,10 +358,19 @@ class GPJob(GPResource):
         self.log_files = self.info['logFiles']
         self.output_files = self.info['outputFiles']
         self.num_output_files = self.info['numOutputFiles']
+        self.input_params = self.info['inputParams']
 
         # Create children, if relevant
         self.children = self.get_child_jobs()
 
+    def get_input_params(self):
+        """Return the input parameters used to launch the job"""
+
+        # Lazily load info
+        if self.info is None: self.get_info()
+
+        return { list(p.keys())[0]:list(p.values())[0] for p in self.input_params }
+
     def get_child_jobs(self):
         """
         Queries the GenePattern server for child jobs of this job, creates GPJob
@@ -522,6 +545,15 @@ class GPJob(GPResource):
         request.add_header('User-Agent', 'GenePatternRest')
         urllib.request.urlopen(request)
 
+    def terminate(self):
+        """Terminate a running or pending job"""
+        url = f'{self.server_data.url}/rest/v1/jobs/{self.job_number}/terminate'
+        request = urllib.request.Request(url, method='DELETE')
+        if self.server_data.authorization_header() is not None:
+            request.add_header('Authorization', self.server_data.authorization_header())
+        request.add_header('User-Agent', 'GenePatternRest')
+        return urllib.request.urlopen(request).code == 200
+
 
 class GPJobSpec(object):
     """
